import ast
from typing import List, cast

from macropy.core.macros import Macros
from macropy.core.quotes import macros, q, ast_literal, name

macros = Macros()  # noqa: F811


@macros.decorator
def monadic(tree, gen_sym, **kw):
    if isinstance(tree, ast.FunctionDef):
        return desugar_statement_list([tree], gen_sym)[0]
    else:
        raise Exception("@monad decorator can only be used for function definitions")

# Transforms one of these:
#
#    def my_function(...):
#        bar = 42
#        import whatever
#        ...
#        bippy = ~(x)
#        [...more statements...]
#
# into one of these:
#    bar = 42
#    import whatever
#    ...
#    def rest(bippy):
#        [...more statements...]


def desugar_statement_list(stmts: List[ast.stmt], gen_sym) -> List[ast.stmt]:
    idx = len(stmts)
    while idx >= 0:
        remainder = stmts[idx:]
        remainder_desugared = desugar_remainder(remainder, gen_sym)
        stmts = stmts[:idx] + remainder_desugared
        idx -= 1
    return stmts


def desugar_remainder(remainder: List[ast.stmt], gen_sym) -> List[ast.stmt]:
    if not remainder:
        return []

    stmt = remainder[0]
    if isinstance(stmt, ast.FunctionDef):
        stmt.body = desugar_statement_list(stmt.body, gen_sym)
        return remainder
    elif isinstance(stmt, ast.AsyncFunctionDef):
        stmt.body = desugar_statement_list(stmt.body, gen_sym)
        return remainder
    elif isinstance(stmt, ast.ClassDef):
        return remainder
    elif isinstance(stmt, ast.Return):
        return remainder
    elif isinstance(stmt, ast.Delete):
        return remainder
    elif isinstance(stmt, ast.Assign):
        assign = cast(ast.Assign, stmt)
        if (isinstance(assign.value, ast.UnaryOp) and isinstance(assign.value.op, ast.Invert)):
            # Input:
            #   foo = ~bar
            #   [...rest...]
            #   return 42
            # Output:
            #   def temp(bar_bind_result):
            #      foo = bar_bind_result  # (handles various assignment stuff)
            #      [...rest...]
            #      return 42
            #   bar.{map, flat_map}(temp)
            unary_op = cast(ast.UnaryOp, assign.value)
            func_name = gen_sym()

            flat_map_arg_name = gen_sym()
            assign.value = q[name[flat_map_arg_name]]

            flat_map_function_def = ast.FunctionDef(
                func_name,
                ast.arguments([ast.arg(flat_map_arg_name, None)], None, [], [], None, []),
                remainder if remainder else [],  # TODO
                [], None)
            flat_map_call = q[ast_literal[unary_op.operand].flat_map(name[func_name])]

            return_call = ast.Return(flat_map_call)

            return [flat_map_function_def, return_call]
        else:
            return remainder
    elif isinstance(stmt, ast.AugAssign):
        return remainder
    elif isinstance(stmt, ast.AnnAssign):
        return remainder
    elif isinstance(stmt, ast.For):
        return remainder
    elif isinstance(stmt, ast.AsyncFor):
        return remainder
    elif isinstance(stmt, ast.While):
        return remainder
    elif isinstance(stmt, ast.If):
        stmt.body = desugar_statement_list(stmt.body, gen_sym)
        stmt.orelse = desugar_statement_list(stmt.orelse, gen_sym)
        return remainder
    elif isinstance(stmt, ast.With):
        stmt.body = desugar_statement_list(stmt.body, gen_sym)
        return remainder
    elif isinstance(stmt, ast.AsyncWith):
        stmt.body = desugar_statement_list(stmt.body, gen_sym)
        return remainder
    elif isinstance(stmt, ast.Raise):
        return remainder
    elif isinstance(stmt, ast.Try):
        stmt.body = desugar_statement_list(stmt.body, gen_sym)
        # NOTE: Semantics are not clear here.  Should try to tie this into
        # ZIO's error types.

        stmt.orelse = desugar_statement_list(stmt.orelse, gen_sym)
        stmt.finalbody = desugar_statement_list(stmt.finalbody, gen_sym)

        for handler in stmt.handlers:
            handler.body = desugar_statement_list(handler.body, gen_sym)
        return remainder
    elif isinstance(stmt, ast.Assert):
        return remainder
    elif isinstance(stmt, ast.Import):
        return remainder
    elif isinstance(stmt, ast.ImportFrom):
        return remainder
    elif isinstance(stmt, ast.Global):
        return remainder
    elif isinstance(stmt, ast.Nonlocal):
        return remainder
    elif isinstance(stmt, ast.Expr):
        expr = cast(ast.Expr, stmt.value)
        if isinstance(expr, ast.BoolOp):
            return remainder
        elif isinstance(expr, ast.BinOp):
            return remainder
        elif isinstance(expr, ast.UnaryOp):
            if isinstance(expr.op, ast.Invert):
                unary_op = stmt.value

                # Input:
                #   ~bar
                #   [...rest...]
                #   return 42
                # Output:
                #   def temp(throwaway):
                #      [...rest...]
                #      return 42
                #   bar.{map, flat_map}(temp)
                func_name = gen_sym()

                flat_map_arg_name = gen_sym()

                flat_map_function_def = ast.FunctionDef(
                    func_name,
                    ast.arguments([ast.arg(flat_map_arg_name, None)], None, [], [], None, []),
                    remainder[1:] if remainder[1:] else [],  # TODO
                    [], None)
                flat_map_call = q[ast_literal[unary_op.operand].flat_map(name[func_name])]
                return_call = ast.Return(flat_map_call)

                return [flat_map_function_def, return_call]
            else:
                return remainder
        elif isinstance(expr, ast.Lambda):
            return remainder
        elif isinstance(expr, ast.IfExp):
            return remainder
        elif isinstance(expr, ast.Dict):
            return remainder
        elif isinstance(expr, ast.Set):
            return remainder
        elif isinstance(expr, ast.ListComp):
            return remainder
        elif isinstance(expr, ast.SetComp):
            return remainder
        elif isinstance(expr, ast.DictComp):
            return remainder
        elif isinstance(expr, ast.GeneratorExp):
            return remainder
        elif isinstance(expr, ast.Await):
            return remainder
        elif isinstance(expr, ast.Yield):
            return remainder
        elif isinstance(expr, ast.YieldFrom):
            return remainder
        elif isinstance(expr, ast.Compare):
            return remainder
        elif isinstance(expr, ast.Call):
            return remainder
        elif isinstance(expr, ast.Compare):
            return remainder
        elif isinstance(expr, ast.Num):
            return remainder
        elif isinstance(expr, ast.Str):
            return remainder
        elif isinstance(expr, ast.FormattedValue):
            return remainder
        elif isinstance(expr, ast.JoinedStr):
            return remainder
        elif isinstance(expr, ast.Bytes):
            return remainder
        elif isinstance(expr, ast.NameConstant):
            return remainder
        elif isinstance(expr, ast.Ellipsis):
            return remainder
        elif isinstance(expr, ast.Attribute):
            return remainder
        elif isinstance(expr, ast.Subscript):
            return remainder
        elif isinstance(expr, ast.Starred):
            return remainder
        elif isinstance(expr, ast.Name):
            return remainder
        elif isinstance(expr, ast.List):
            return remainder
        elif isinstance(expr, ast.Tuple):
            return remainder
        else:
            return remainder
    elif isinstance(stmt, ast.Pass):
        return remainder
    elif isinstance(stmt, ast.Break):
        return remainder
    elif isinstance(stmt, ast.Continue):
        return remainder
    else:
        raise TypeError(f"Unexpected ast statement type: {stmt}")



from dataclasses import dataclass
from typing import Any, Callable, Generic, NoReturn, TypeVar

from zio_py.cause import Cause

A = TypeVar('A', covariant=True)
E = TypeVar('E', covariant=True)
R = TypeVar('R', contravariant=True)

B = TypeVar('B')

# Some invariant analogues of the above types
A0 = TypeVar('A0')
A1 = TypeVar('A1')
E1 = TypeVar('E1')
R1 = TypeVar('R1')


class ZIO(Generic[R, E, A]):
    def flat_map(self: 'ZIO[R1, E1, A]', k: Callable[[A], 'ZIO[R1, E1, B]']) -> 'ZIO[R1, E1, B]':
        return FlatMap(self, k)

    def fold(self, failure: Callable[[E], B], success: Callable[[A], B]) -> 'ZIO[R, NoReturn, B]':
        x: MapFn[R, NoReturn, E, B] = MapFn(failure)
        y: MapFn[R, NoReturn, A, B] = MapFn(success)
        return self.fold_m(x, y)

    def fold_m(self, failure: Callable[[E], 'ZIO[R, E1, B]'],
               success: Callable[[A], 'ZIO[R, E1, B]']) -> 'ZIO[R, E1, B]':
        return self.fold_cause_m(FoldCauseMFailureFn(failure), success)

    def fold_cause_m(self, failure: Callable[[Cause[E]], 'ZIO[R, E1, B]'],
                     success: Callable[[A], 'ZIO[R, E1, B]']) -> 'ZIO[R, E1, B]':
        return Fold(self, failure, success)

    def map(self, f: Callable[[A], B]) -> 'ZIO[R, E, B]':
        g = self._map_helper(f)
        return FlatMap(self, g)

    def _map_helper(self, f: Callable[[A1], B]) -> Callable[[A1], 'ZIO[R, E, B]']:
        return lambda x: ZIOStatic.succeed(f(x))

    def provide(self, r: R) -> 'IO[E, A]':
        return ZIOStatic.provide(r)(self)

    def __invert__(self) -> A:
        pass

# NOTE: Putting these static methods in a different class to avoid name clashes.
# It would be nice to put them in the ZIO class, but...Python.


class ZIOStatic:
    @staticmethod
    def access(f: Callable[[R1], A1]) -> 'ZIO[R1, NoReturn, A1]':
        return Read(lambda r: ZIOStatic.succeed(f(r)))

    @staticmethod
    def access_m(f: Callable[[R], 'ZIO[R, E, A]']) -> 'ZIO[R, E, A]':
        return Read(f)

    @staticmethod
    def effect(effect: Callable[[], A]) -> 'Task[A]':
        return EffectPartial(effect)

    @staticmethod
    def effect_total(effect: Callable[[], A]) -> 'UIO[A]':
        return EffectTotal(effect)

    @staticmethod
    def environment() -> ZIO[R1, NoReturn, R1]:
        return ZIOStatic.access(lambda r: r)

    @staticmethod
    def halt(cause: Cause[E1]) -> 'IO[E1, NoReturn]':
        return Fail(lambda: cause)

    @staticmethod
    def provide(r: R1) -> Callable[['ZIO[R1, E1, A1]'], 'IO[E1, A1]']:
        return lambda zio: Provide(r, zio)

    @staticmethod
    def succeed(a: A1) -> 'UIO[A1]':
        return Succeed(a)


RIO = ZIO[R, Exception, A]
URIO = ZIO[R, NoReturn, A]
IO = ZIO[Any, E, A]
UIO = ZIO[Any, NoReturn, A]
Task = ZIO[Any, Exception, A]


class EffectPartial(Generic[A], Task[A]):
    def __init__(self, effect: Callable[[], A]) -> None:
        self.effect = effect


class EffectTotal(Generic[A], UIO[A]):
    def __init__(self, effect: Callable[[], A]) -> None:
        self.effect = effect


@dataclass(frozen=True)
class Fail(Generic[E, A], IO[E, A]):
    cause: Callable[[], Cause[E]]


class FlatMap(Generic[R, E1, A0, A1], ZIO[R, E1, A1]):
    def __init__(self, zio: ZIO[R, E1, A0], k: Callable[[A0], ZIO[R, E1, A1]]) -> None:
        self._zio = zio
        self._k = k

    @property
    def zio(self) -> ZIO[R, E1, A0]:
        return self._zio

    @property
    def k(self) -> Callable[[A0], ZIO[R, E1, A1]]:
        return self._k


class Fold(Generic[R, E, E1, A1, B], ZIO[R, E1, B]):
    def __init__(
        self,
        value: ZIO[R, E, A1],
        failure: Callable[[Cause[E]], ZIO[R, E1, B]],
        success: Callable[[A1], ZIO[R, E1, B]]
    ) -> None:
        self._value = value
        self._failure = failure
        self._success = success

    @property
    def value(self) -> ZIO[R, E, A1]:
        return self._value

    @property
    def failure(self) -> Callable[[Cause[E]], ZIO[R, E1, B]]:
        return self._failure

    @property
    def success(self) -> Callable[[A1], ZIO[R, E1, B]]:
        return self._success

    def __call__(self, v: A1) -> ZIO[R, E1, B]:
        return self.success(v)


class FoldCauseMFailureFn(Generic[R, E, E1, A]):
    def __init__(self, underlying: Callable[[E], ZIO[R, E1, A]]):
        self._underlying = underlying

    def __call__(self, c: Cause[E]) -> ZIO[R, E1, A]:
        return c.failure_or_cause().fold(self._underlying, ZIOStatic.halt)


class MapFn(Generic[R1, E1, A1, B], ZIO[R1, E1, B]):
    def __init__(self, underlying: Callable[[A1], B]) -> None:
        self._underlying = underlying

    def __call__(self, a: A1) -> ZIO[R1, E1, B]:
        return Succeed(self._underlying(a))


@dataclass(frozen=True)
class Provide(Generic[R, E, A], IO[E, A]):
    r: R
    next: ZIO[R, E, A]


class Read(Generic[R, E, A], ZIO[R, E, A]):
    def __init__(self, k: Callable[[R], ZIO[R, E, A]]):
        self._k = k

    @property
    def k(self) -> Callable[[R], ZIO[R, E, A]]:
        return self._k


@dataclass(frozen=True)
class Succeed(Generic[A], UIO[A]):
    value: A



from typing import Any, Generic, List, TypeVar

from zio_py.cause import Cause
from zio_py.zio import (ZIO, EffectPartial, EffectTotal, Fail, FlatMap, Fold,
                        MapFn, Provide, Read, Succeed)

A = TypeVar('A')
E = TypeVar('E')
R = TypeVar('R', covariant=True)


class Runtime(Generic[R]):
    def __init__(self) -> None:
        self.environments: List[Any] = []

    def unsafe_run_sync(self: 'Runtime[R]', zio: ZIO[Any, E, A]) -> A:
        if isinstance(zio, FlatMap):
            k = zio.k
            a = self.unsafe_run_sync(zio.zio)
            return self.unsafe_run_sync(k(a))  # type: ignore
        elif isinstance(zio, EffectTotal):
            return zio.effect()
        elif isinstance(zio, EffectPartial):
            return zio.effect()  # type: ignore
        elif isinstance(zio, Fail):
            raise zio.cause.value  # type: ignore
        elif isinstance(zio, MapFn):
            return zio(a)  # type: ignore
        elif isinstance(zio, Fold):
            value = self.unsafe_run_sync(zio.value)
            if isinstance(value, Cause):
                return self.unsafe_run_sync(zio.failure(value))
            else:
                return self.unsafe_run_sync(zio.success(value))
        elif isinstance(zio, Provide):
            self.environments.append(zio.r)
            # TODO: Bracket the environment
            return self.unsafe_run_sync(zio.next)  # type: ignore
        elif isinstance(zio, Read):
            r = self.environments[-1]
            return self.unsafe_run_sync(zio.k(r))  # type: ignore
        elif isinstance(zio, Succeed):
            return zio.value
        else:
            raise ValueError(f"Unexpected instruction: {zio}")



from zio_py.cause import Cause, Die
from zio_py.console import Console, LiveConsole, read_line, print_line
from zio_py.either import Either
from zio_py.runtime import Runtime
import zio_py.syntax
from zio_py.zio import ZIO, ZIOStatic



from abc import ABCMeta
from typing import Callable, Generic, TypeVar

A = TypeVar('A', covariant=True)
B = TypeVar('B', covariant=True)
C = TypeVar('C', covariant=True)


class Either(Generic[A, B], metaclass=ABCMeta):
    def fold(self, left: Callable[[A], C], right: Callable[[B], C]) -> C:
        # TODO
        pass



from distutils.core import setup

setup(
    name='zio_py',
    packages=['zio_py'],
    version='0.0.4a',
    license='Apache license 2.0',
    description='Python port of Scala ZIO for pure functional programming',
    long_description="""ZIO for Python (see https://github.com/zio/zio)

This is a fairly straightforward (unofficial and plucky) port of the
Scala ZIO library.  Work in progress.
    """,
    author='William Harvey',
    author_email='drwjharvey@gmail.com',
    url='https://github.com/harveywi/ziopy',
    download_url='https://github.com/harveywi/ziopy/archive/0.0.4a.tar.gz',
    keywords=['ZIO', 'IO', 'monads', 'pure fp', 'functional programming',
                'monad syntax'],
    install_requires=[
        'macropy3==1.1.0b2'
    ],
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7'
    ]
)



from abc import ABCMeta, abstractmethod
from typing import NoReturn, TypeVar

from zio_py.zio import ZIO, ZIOStatic

R = TypeVar('R')


class Console(metaclass=ABCMeta):
    @abstractmethod
    def print(self, line: str) -> ZIO[R, NoReturn, None]:
        pass

    @abstractmethod
    def input(self, prompt: str) -> ZIO[R, Exception, str]:
        pass


class LiveConsole(Console):
    def print(self, line: str) -> ZIO[R, NoReturn, None]:
        return ZIOStatic.effect_total(lambda: print(line))

    def input(self, prompt: str) -> ZIO[R, Exception, str]:
        return ZIOStatic.effect(lambda: input(prompt))


def print_line(line: str) -> ZIO[Console, NoReturn, None]:
    return ZIOStatic.access_m(lambda env: env.print(line))


def read_line(prompt: str) -> ZIO[Console, Exception, str]:
    return ZIOStatic.access_m(lambda env: env.input(prompt))



from abc import ABCMeta, abstractmethod
from typing import Generic, NoReturn, TypeVar

from zio_py.either import Either

E = TypeVar('E', covariant=True)


class Cause(Generic[E], metaclass=ABCMeta):
    @abstractmethod
    def failure_or_cause(self) -> Either[E, 'Cause[NoReturn]']:
        # TODO
        pass


class Die(Cause[NoReturn]):
    value: Exception
